@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2017 PLT Design Inc.
+  Copyright (c) 2004-2018 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -64,10 +64,12 @@ static Scheme_Object *unsafe_vector_set (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_vector_star_len (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_vector_star_ref (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_vector_star_set (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_vector_star_cas (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_struct_ref (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_struct_set (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_struct_star_ref (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_struct_star_set (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_struct_star_cas (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_string_len (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_string_ref (int argc, Scheme_Object *argv[]);
 static Scheme_Object *unsafe_string_set (int argc, Scheme_Object *argv[]);
@@ -133,6 +135,12 @@ scheme_init_vector (Scheme_Env *env)
   SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_NARY_INLINED);
   scheme_add_global_constant("vector-set!", p, env);
 
+  p = scheme_make_noncm_prim(scheme_checked_vector_cas,
+                             "vector-cas!",
+                             4, 4);
+  SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_NARY_INLINED);
+  scheme_add_global_constant("vector-cas!", p, env);
+
   scheme_add_global_constant("vector->list", 
 			     scheme_make_immed_prim(vector_to_list, 
 						    "vector->list", 
@@ -232,6 +240,10 @@ scheme_init_unsafe_vector (Scheme_Env *env)
   SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_NARY_INLINED);
   scheme_add_global_constant("unsafe-vector*-set!", p, env);  
 
+  p = scheme_make_immed_prim(unsafe_vector_star_cas, "unsafe-vector*-cas!", 4, 4);
+  SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_NARY_INLINED);
+  scheme_add_global_constant("unsafe-vector*-cas!", p, env);
+
   REGISTER_SO(scheme_unsafe_struct_ref_proc);
   p = scheme_make_immed_prim(unsafe_struct_ref, "unsafe-struct-ref", 2, 2);
   scheme_unsafe_struct_ref_proc = p;
@@ -256,6 +268,10 @@ scheme_init_unsafe_vector (Scheme_Env *env)
   SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_NARY_INLINED);
   scheme_add_global_constant("unsafe-struct*-set!", p, env);  
 
+  p = scheme_make_immed_prim(unsafe_struct_star_cas, "unsafe-struct*-cas!", 4, 4);
+  SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_NARY_INLINED);
+  scheme_add_global_constant("unsafe-struct*-cas!", p, env);
+
   REGISTER_SO(scheme_unsafe_string_length_proc);
   p = scheme_make_immed_prim(unsafe_string_len, "unsafe-string-length", 1, 1);
   SCHEME_PRIM_PROC_FLAGS(p) |= scheme_intern_prim_opt_flags(SCHEME_PRIM_IS_UNARY_INLINED
@@ -633,6 +649,25 @@ scheme_checked_vector_set(int argc, Scheme_Object *argv[])
   return scheme_void;
 }
 
+Scheme_Object *
+scheme_checked_vector_cas(int argc, Scheme_Object *argv[])
+{
+  Scheme_Object *vec = argv[0];
+  intptr_t i, len;
+
+  if (!SCHEME_MUTABLE_VECTORP(vec))
+    scheme_wrong_contract("vector-cas!", "(and/c vector? (not/c immutable?) (not/c impersonator?))", 0, argc, argv);
+
+  len = SCHEME_VEC_SIZE(vec);
+
+  i = scheme_extract_index("vector-cas!", 1, argc, argv, len, 0);
+
+  if (i >= len)
+    return bad_index("vector-cas!", "", argv[1], argv[0], 0);
+
+  return unsafe_vector_star_cas(argc, argv);
+}
+
 # define cons(car, cdr) scheme_make_pair(car, cdr)
 
 Scheme_Object *
@@ -929,7 +964,7 @@ static Scheme_Object *do_chaperone_vector(const char *name, int is_impersonator,
   Scheme_Chaperone *px;
   Scheme_Object *val = argv[0];
   Scheme_Object *redirects;
-  Scheme_Hash_Tree *props;
+  Scheme_Object *props;
 
   if (SCHEME_CHAPERONEP(val)) {
     val = SCHEME_CHAPERONE_VAL(val);
@@ -1076,6 +1111,28 @@ static Scheme_Object *unsafe_vector_star_set (int argc, Scheme_Object *argv[])
   return scheme_void;
 }
 
+static Scheme_Object *unsafe_vector_star_cas (int argc, Scheme_Object *argv[])
+{
+  Scheme_Object *vec = argv[0];
+  Scheme_Object *idx = argv[1];
+  Scheme_Object *ov = argv[2];
+  Scheme_Object *nv = argv[3];
+
+#ifdef MZ_USE_FUTURES
+  return mzrt_cas((volatile uintptr_t *)(SCHEME_VEC_ELS(vec) + SCHEME_INT_VAL(idx)),
+                  (uintptr_t)ov, (uintptr_t)nv)
+    ? scheme_true : scheme_false;
+#else
+  /* For cooperative threading, no atomicity required */
+  if (SCHEME_VEC_ELS(vec)[SCHEME_INT_VAL(idx)] == ov) {
+    SCHEME_VEC_ELS(vec)[SCHEME_INT_VAL(idx)] = nv;
+    return scheme_true;
+  } else {
+    return scheme_false;
+  }
+#endif
+}
+
 static Scheme_Object *unsafe_struct_ref (int argc, Scheme_Object *argv[])
 {
   if (SCHEME_CHAPERONEP(argv[0]))
@@ -1104,6 +1161,28 @@ static Scheme_Object *unsafe_struct_star_set (int argc, Scheme_Object *argv[])
   return scheme_void;
 }
 
+static Scheme_Object *unsafe_struct_star_cas (int argc, Scheme_Object *argv[])
+{
+  Scheme_Object *s = argv[0];
+  Scheme_Object *idx = argv[1];
+  Scheme_Object *ov = argv[2];
+  Scheme_Object *nv = argv[3];
+
+#ifdef MZ_USE_FUTURES
+  return (mzrt_cas((volatile uintptr_t *)(&((Scheme_Structure *)s)->slots[SCHEME_INT_VAL(idx)]),
+                   (uintptr_t)ov, (uintptr_t)nv)
+          ? scheme_true : scheme_false);
+#else
+  /* For cooperative threading, no atomicity required */
+  if (((Scheme_Structure *)s)->slots[SCHEME_INT_VAL(idx)] == ov) {
+    ((Scheme_Structure *)s)->slots[SCHEME_INT_VAL(idx)] = nv;
+    return scheme_true;
+  } else {
+    return scheme_false;
+  }
+#endif
+}
+
 static Scheme_Object *unsafe_string_len (int argc, Scheme_Object *argv[])
 {
   intptr_t n = SCHEME_CHAR_STRLEN_VAL(argv[0]);
